#!/usr/bin/env python
"""
A little and artificial example to show how to properly and efficiently
do cross-server upload joins.

get_basic_data is of course a silly function.

remote_match, on the other hand, probably is a good starting point for a
more general functionality.

In real life, you'd have a much larger match_size (1e7 ought to be possible
depending on several details), and you probably need to use run_async
rather than run_sync, but that's about it.

This assumes there's enough RAM for the full match; if that assumption is
not true, you either need to get a computer manufactured in this millenium
or re-think your problem.
"""

from astropy import table
import pyvo


def iter_slices(total_length, batch_size):
  """iterates over slices of up to batch_size filling 0 to total_length.
  """
  limits = list(range(0, total_length, batch_size))+[batch_size]
  for lower, upper in zip(limits[:-1], limits[1:]):
    if lower<upper:
      yield slice(lower, upper)


def remote_match(svc, source_table, remote_table, batch_size, match_column):
  """adds records from remote_table on svc to source_table.
  """
  matched_records = []
  match_on = source_table[match_column]

  # only match the match_column (for a positional crossmatch, use
  # an id column (create one if necessary) and the positions).
  for slice in iter_slices(len(source_table), batch_size):
    result = svc.run_sync("""SELECT a.* FROM
      {remote_table} AS a JOIN
      TAP_UPLOAD.mine AS b
      USING ({match_column})""".format(**locals()), 
      # the next line is where most of the magic is.
      uploads={"mine": table.Table([match_on[slice]])})
    matched_records.append(result.to_table())
  
  joined_match = table.vstack(matched_records)
  del matched_records

  # now fiddle back what we've pulled from the server into the source_table.
  return table.join(source_table, joined_match, keys=match_column)


def get_basic_data(svc):
  """returns some test data from svc.

  Here, that's a bunch of full records from a Gaia service.

  Note: If you're doing SELECT *, you're almost certainly doing it wrong.
  In actual applications, think what columns you actually need.
  """
  result = svc.run_sync("""
    SELECT TOP 400 * 
    FROM gaiaedr3.gaia_source
    WHERE 
      source_id BETWEEN 4657847914607935488 AND 4657988652096290815
    """)
  return result.to_table()


def main():
  my_gaia_part = get_basic_data(
    pyvo.dal.TAPService("https://gea.esac.esa.int/tap-server/tap"))
  with_remote_data = remote_match(
    pyvo.dal.TAPService("http://dc.g-vo.org/tap"),
    source_table=my_gaia_part,
    remote_table="gedr3spur.main",
    batch_size=100,
    match_column="source_id")

  with open("matched_stuff.vot", "wb") as f:
    with_remote_data.write(output=f, format="votable")


if __name__=="__main__":
  main()
